import asyncio
import random
import logging


class Arduino(object):
    """ Utility functions to test Arduino sensor read outside of RoV """
    def __init__(self, serial_port, baud_rate=115200):
        self.logger = logging.getLogger("Arduino")
        random.seed()
        # Set some reasonable default values
        self.sensors = {
                   'tempWater': 30.0,
                   'pressureWater': 100.0,
                   'batteryVoltage': 12.2
               }

        self.logger.debug(f"Opening mock serial port on {serial_port} at {baud_rate} baud.")

    async def get_sensors(self):
        self.sensors['tempWater'] += random.uniform(-.1, .1)
        self.sensors['pressureWater'] += random.uniform(-.1, .1)
        self.sensors['batteryVoltage'] += random.uniform(-.1, .1)
        await asyncio.sleep(0.2)
        return self.sensors

    async def set_actuators(self, values):
        vertical  = int(round(100 * values["vertical"]))
        starboard = int(round(100 * values["starboard"]))
        port      = int(round(100 * values["port"]))
        lights    = int(round(values['lights']))

        self.logger.debug(f"vertical={vertical};starboard={starboard};port={port};lights={lights}".encode('ascii'))
        await asyncio.sleep(.2)



import argparse
import logging
import shutil
from pathlib import Path
import os

__version__ = "0.0.9"

from .server.cameraserver import CameraServer
from .server.ioserver import IOServer
from .server.advertising import AdvertisingServer
    
def setup_startup():
    directory = Path(__file__).parent
    systemd_user_dir = os.path.expanduser("~/.config/systemd/user/")

    os.makedirs(systemd_user_dir)

    shutil.copy(directory / "pyedurov2.service", systemd_user_dir)

    os.system("systemctl --user daemon-reload")
    os.system("systemctl --user enable pyedurov2.service")

    print("pyedurov2 service added to startup.")



def edurov_web():
    parser = argparse.ArgumentParser(
        description='Start the Engage eduROV web server')
    parser.add_argument(
        '-r',
        metavar='RESOLUTION',
        type=str,
        default='1280x720',
        help='''resolution, use format WIDTHxHEIGHT (default 1280x720)''')
    parser.add_argument(
        '-fps',
        metavar='FRAMERATE',
        type=int,
        default=30,
        help='framerate for the camera (default 30)')
    parser.add_argument(
        '-port',
        metavar='SERVER_PORT',
        type=int,
        default=8080,
        help="which port the server should serve it's main page (default 80)")
    parser.add_argument(
        '-serial',
        metavar='SERIAL_PORT',
        type=str,
        default=None,
        help="which serial port the script should try to use to communicate with the Arduino module")
    parser.add_argument(
        '--loglevel',
        type=str,
        default='INFO',
        help='Set log level')
    parser.add_argument(
        '--runatstartup',
        action='store_true',
        help='Setup server to run at startup and exit program.')

    args = parser.parse_args()

    if args.runatstartup:
        print("Setting up pyedurov2 to execute at startup.")
        setup_startup()
        return

    logging.basicConfig(level=args.loglevel)

    ioserver = IOServer(args.serial, loglevel=args.loglevel)
    camera = CameraServer(args.r, args.fps, args.loglevel)
    advertisingserver = AdvertisingServer(loglevel=args.loglevel)

    logging.info("Waiting for websocket servers to go online before starting web server")
    ioserver.ready.wait()
    camera.ready.wait()
    advertisingserver.ready.wait()

    ioserver.join()
    logging.info("I/O server stopped. Killing other services and exiting program.")
    camera.stop()
    advertisingserver.stop()
    
if __name__ == '__main__':
    edurov_web()


import logging
import psutil
import sense_hat


class SenseHat(object):
    """ Utility functions to read Sense Hat data asynchronously """
    def __init__(self):
        self.logger = logging.getLogger("SenseHat")
        self.sense = None
        try:
            self.sense = sense_hat.SenseHat()
            self.sense.show_message("Hello World")
        except Exception as e:
            self.logger.warning("Sense hat is not available",  exc_info=True)

    async def get_sensors(self):
        if self.sense is None:
            return dict()
        
        orientation = self.sense.get_orientation()
        return {
                    'temp':     self.sense.get_temperature(),
                    'pressure': self.sense.get_pressure() / 10,
                    'humidity': self.sense.get_humidity(),
                    'pitch':    orientation['pitch'],
                    'roll':     orientation['roll'],
                    'yaw':      orientation['yaw']
                }


class SystemMonitor(object):
    """ Utility functions to read Raspberry pi system state and Sense Hat data asynchronously """
    def __init__(self):
        self.logger = logging.getLogger("SystemMonitor")

    async def get_system_data(self):
        temps = psutil.sensors_temperatures()
        free_disk = psutil.disk_usage("/").free / 1024 / 1024
        cpu_usage = psutil.cpu_percent() 
        try:
            temp = list(temps.values())[0][0].current
        except:
            temp = "N/A"
        
        return {
                    'free_space': free_disk,
                    'cpu_temp': temp,
                    'cpu_load': cpu_usage
                }



import logging
import threading
import cv2

from utility import StreamingOutput


class Camera(object):
    def __init__(self, video_resolution, fps):
        self.logger = logging.getLogger("Camera")
        self.fps = fps
        self.stream = StreamingOutput()
        self.resolution = video_resolution
        self.capture = None
        self.timer = None

    def open(self):
        try:
            self.capture = cv2.VideoCapture(0)
            self.capture.set(cv2.CAP_PROP_FRAME_WIDTH, self.resolution[0])
            self.capture.set(cv2.CAP_PROP_FRAME_HEIGHT, self.resolution[1])
            self.capture.set(cv2.CAP_PROP_SATURATION, 0.2)
            self.capture_frame()
        except Exception as e:
            self.logger.warning(e)
            self.logger.warning("Camera initialization failed, camera is not available")
            self.capture = None

    def close(self):
        if self.timer is not None:
            self.timer.cancel()
            self.timer.join()
            self.timer = None
        if self.capture is not None:
            self.capture.release()
            self.capture = None

    def capture_frame(self):
        self.timer = threading.Timer(1 / self.fps, self.capture_frame)
        self.timer.start()
        try:
            rc, img = self.capture.read()
            if not rc:
                return

            rc, buffer = cv2.imencode('.jpg', img)
            if not rc:
                return
            self.stream.write(buffer.tobytes())
        except KeyboardInterrupt:
            self.timer.cancel()

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()




# Import mock interfaces if not running on raspberry pi.
from ..utility import is_raspberrypi

if not is_raspberrypi():
    from .mock import system as system
    from .mock import camera as camera
else:
    from . import system as system
    from . import camera as camera

from . import arduino as arduino



import json
import logging
import multiprocessing
import asyncio
import time

import websockets as websockets

from ..hardware import arduino
from ..hardware import system
from ..utility import get_host_ip


class IOServer(multiprocessing.Process):
    """ Creates a new process that Exposes the Arduino, sense hat, and system sensors and actuators as a websocket data stream """
    def __init__(self, arduino_serial_port=None, arduino_baud_rate=115200, loglevel="INFO", port=8082):
        self.port = port
        self.serial_port = arduino_serial_port
        self.baud_rate = arduino_baud_rate
        self.loglevel = loglevel
        self.sensors = dict()
        self.actuators = dict()
        self.start_time = time.time()
        self.sensor_queue = None
        self.actuator_queue = None
        self.server = None
        self.ready = multiprocessing.Event()
        self.stop_event = multiprocessing.Event()

        super().__init__(target=self._runner, daemon=True)
        self.start()

    async def stop(self):
        self.stop_event.set()
        self.join()
        logging.debug("I/O process terminated")
    
    async def _collect_arduino_sensors(self):
        while not self.stop_event.is_set():
            # Don't get arduino data if we're not serving clients.
            if self.server.ws_server.websockets:
                self.logger.debug("Get arduino data")
                await self.sensor_queue.put(await self.arduino.get_sensors())
                await self.sensor_queue.join()
            else:
                await asyncio.sleep(1)

    async def _collect_sense_hat_sensors(self):
        raspberry = system.SenseHat()
        while not self.stop_event.is_set():
            # Don't get arduino data if we're not serving clients.
            if self.server.ws_server.websockets:
                self.logger.debug("Get sense hat data")
                await self.sensor_queue.put(await raspberry.get_sensors())
                await self.sensor_queue.join()
            await asyncio.sleep(0.3)

    async def _collect_system_data(self):
        raspberry = system.SystemMonitor()
        while not self.stop_event.is_set():
            # Don't get arduino data if we're not serving clients.
            if self.server.ws_server.websockets:
                self.logger.debug("Get system data")
                await self.sensor_queue.put(await raspberry.get_system_data())
                await self.sensor_queue.join()
            await asyncio.sleep(1)

    async def _send_sensors(self, websocket):
        while not self.stop_event.is_set():
            # empty the sensor queue before sending the data update.
            while not self.stop_event.is_set():
                self.sensors.update(await self.sensor_queue.get())
                empty = self.sensor_queue.empty()
                self.sensor_queue.task_done()
                if empty:
                    break
            
            self.logger.debug("Send arduino data to client")
            await websocket.send(json.dumps(self.sensors))

        # Drain the sensor queue so the other tasks can exit as well.
        while not self.sensor_queue.empty():
            self.sensors.update(await self.sensor_queue.get())
            self.sensor_queue.task_done()

    async def _receive_input(self, websocket):
        while not self.stop_event.is_set():
            self.logger.debug("Wait for incoming websocket data")
            data = await websocket.recv()
            data = json.loads(data)
            self.logger.debug(f"Got websocket data {data}")
            async with self.actuator_lock:
                self.actuators.update(data)
                self.arduino.set_actuators(self.actuators)

    async def _handler(self, websocket, path):
        self.logger.info(f"I/O server received connection from {path}")
        await websocket.recv()
        send = asyncio.create_task(self._send_sensors(websocket))
        receive = asyncio.create_task(self._receive_input(websocket))
        await websocket.wait_closed()
        await send
        await receive

    def _runner(self):
        asyncio.run(self._task())

    async def _task(self):
        logging.basicConfig(level=self.loglevel)
        self.logger = logging.getLogger("I/OServer")
        self.sensor_condition = asyncio.Condition()
        self.actuator_lock = asyncio.Lock()
        self.sensor_queue = asyncio.Queue()
        self.actuator_queue = asyncio.Queue()

        self.server = websockets.serve(self._handler, get_host_ip(), self.port)
        self.logger.info(f"I/O websocket server started at ws://{get_host_ip()}:{self.port}")

        async with arduino.Arduino(self.serial_port, self.baud_rate) as self.arduino:
            arduino_task = asyncio.create_task(self._collect_arduino_sensors())
            hat_task = asyncio.create_task(self._collect_sense_hat_sensors())
            system_task = asyncio.create_task(self._collect_system_data())
            self.logger.debug("Ready to send sensor data")

            self.ready.set()

            await self.server
            await arduino_task
            await hat_task
            await system_task

        self.logger.info('Shutting down server')
        finish = time.time()
        seconds = finish - self.start_time
        self.logger.debug(f'Server was live for {seconds:.1f} seconds')



import logging
import multiprocessing
import asyncio
import time
import socket

from .. import __version__


class AdvertisingServer(multiprocessing.Process):
    """ Creates a new process that Advertises the Edurov server as a UDP multicast beacon """
    def __init__(self, loglevel="INFO", port=8083):
        self.port = port
        self.loglevel = loglevel
        self.start_time = time.time()
        self.server = None
        self.ready = multiprocessing.Event()
        self.stop_event = multiprocessing.Event()

        super().__init__(target=self._runner, daemon=True)
        self.start()

    async def stop(self):
        self.stop_event.set()
        self.join()
        logging.debug("I/O process terminated")

    async def _wait_for_stop_event(self):
        while not self.stop_event.is_set():
            await asyncio.sleep(2)
        self.server.ws_server.close()

    def _runner(self):
        asyncio.run(self._task())

    async def _task(self):
        logging.basicConfig(level=self.loglevel)
        self.logger = logging.getLogger("AdvertisingServer")
                    
        MCAST_GRP = '224.0.0.50'
        MULTICAST_TTL = 1

        self.logger.info(f"Advertising server started at {MCAST_GRP}:{self.port}")
        self.ready.set()

        msg = f"edurov server v{__version__}"

        while not self.stop_event.is_set():
            with socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP) as sock:
                sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, MULTICAST_TTL)
                
                sock.sendto(msg.encode("ascii"), (MCAST_GRP, self.port))

            await asyncio.sleep(2)

        self.logger.info('Shutting down server')
        finish = time.time()
        seconds = finish - self.start_time
        self.logger.debug(f'Server was live for {seconds:.1f} seconds')



import logging
import picamera

from ..utility import StreamingOutput


class Camera(object):
    def __init__(self, video_resolution, fps):
        self.logger = logging.getLogger("Camera")
        self.stream = StreamingOutput()
        try:
            self.camera = picamera.PiCamera(resolution=video_resolution, framerate=fps)
            self.camera.start_recording(self.stream, format='mjpeg')
        except Exception as e:
            self.logger.warning(e)
            self.logger.warning("Camera initialization failed, camera is not available")
            self.camera = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.camera is not None:
            self.camera.close()
